use std::{
    collections::HashMap,
    sync::{
        Arc,
        mpsc::{Receiver, RecvTimeoutError, Sender},
    },
    thread::JoinHandle,
    time::Duration,
};

/// Status receiver of a dead man switch
pub struct DeadManReceiver {
    rx: Receiver<bool>,
}

impl DeadManReceiver {
    /// Checks if a dead man switch is alive
    pub fn alive(&self) -> bool {
        let res = self.rx.recv_timeout(Duration::from_millis(400));
        if let Err(err) = res {
            return match err {
                RecvTimeoutError::Timeout => true,
                RecvTimeoutError::Disconnected => false,
            };
        }

        false
    }
}

/// A Dead Man Switch
pub struct DeadManSwitch {
    pub lifeline: Sender<bool>,
}

impl DeadManSwitch {
    /// Create a new Dead Man Switch
    ///
    /// The Dead Man Switch will notify the receiver and mark itself dead after it got dropped (for example in a panic)
    pub fn new() -> (Self, DeadManReceiver) {
        let (tx, rx) = std::sync::mpsc::channel();

        (Self { lifeline: tx }, DeadManReceiver { rx })
    }
}

impl Drop for DeadManSwitch {
    fn drop(&mut self) {
        log::info!("I am DEAD man 💀");
        let _ = self.lifeline.send(true);
    }
}

/// Service Manager
///
/// This will manage and keep background services running-
pub struct ServiceManager {
    pub services: HashMap<String, Arc<Box<dyn Fn(DeadManSwitch) -> () + 'static + Send + Sync>>>,
    pub mode: ServiceMode,
}

// TODO : impl decay mode

/// The mode on which services should operate
pub enum ServiceMode {
    /// Behave like a daemon. Services can never die and will come back to life after beeing killed. They will always haunt you.
    Daemon,
    /// Graceful Decay. Services stay dead after beeing killed. If no services are running anymore the `ServiceManager` will commit suicide.
    Decay,
}

impl ServiceManager {
    pub fn new() -> Self {
        Self {
            services: HashMap::new(),
            mode: ServiceMode::Daemon,
        }
    }

    /// Set a `ServiceMode` to control the behaviour of services.
    pub fn mode(mut self, mode: ServiceMode) -> Self {
        self.mode = mode;
        self
    }

    /// Register a new background service
    pub fn register<T: Fn(DeadManSwitch) -> () + 'static + Send + Sync>(
        mut self,
        service_name: &str,
        service: T,
    ) -> Self {
        self.services
            .insert(service_name.to_string(), Arc::new(Box::new(service)));
        self
    }

    fn run(self) {
        let mut threads = HashMap::new();

        // Initial start
        for (service, f) in self.services.iter() {
            log::info!("Starting service {service}");
            let (dms, dmr) = DeadManSwitch::new();
            let f = Arc::clone(&f);
            threads.insert(
                service.clone(),
                (
                    std::thread::spawn(move || {
                        f(dms);
                    }),
                    dmr,
                ),
            );
        }

        // Monitor loop
        loop {
            let mut threads_new = HashMap::new();
            let mut threads_rm = Vec::new();

            for (service, (t, dm)) in &mut threads {
                let alive = dm.alive();
                log::trace!(
                    "Checking up on {service} -- {}",
                    if alive { "ALIVE" } else { "DEAD" }
                );

                if !alive {
                    // Restart (Daemon)
                    match self.mode {
                        ServiceMode::Daemon => {
                            log::warn!("Service {service} died. Restarting...");

                            while !t.is_finished() {
                                log::debug!("waiting for thread");
                            }

                            let f = Arc::clone(self.services.get(service).unwrap());

                            let (dms, dmr) = DeadManSwitch::new();
                            threads_new.insert(
                                service.clone(),
                                (
                                    std::thread::spawn(move || {
                                        f(dms);
                                    }),
                                    dmr,
                                ),
                            );
                        }
                        ServiceMode::Decay => {
                            log::info!("Service {service} died.");
                            threads_rm.push(service.clone());
                        }
                    }
                }
            }

            for t in threads_rm {
                threads.remove(&t);
            }

            threads.extend(threads_new);

            if threads.is_empty() && matches!(self.mode, ServiceMode::Decay) {
                log::info!("All my services are dead. Commiting suicide...");
                break;
            }
        }
    }

    /// Start the services
    ///
    /// This returns a ThreadHandle for the service monitor thread.
    pub fn spawn(self) -> JoinHandle<()> {
        std::thread::spawn(|| self.run())
    }
}
